/*
 * Copyright 2011 Apaq.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package dk.apaq.crud.vaadin;

import com.vaadin.data.Property.ValueChangeEvent;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;


import com.vaadin.data.Buffered;
import com.vaadin.data.Item;
import com.vaadin.data.Item.PropertySetChangeNotifier;
import com.vaadin.data.Property;
import com.vaadin.data.Property.ValueChangeListener;
import com.vaadin.data.Validator.InvalidValueException;
import com.vaadin.data.util.BeanItem;
import dk.apaq.crud.Crud;
import dk.apaq.crud.Crud.Editable;
import dk.apaq.crud.HasId;


/**
 * An
 */
public class CrudItem<IDT, BT> implements HasBean<BT>, Item, Buffered,
        PropertySetChangeNotifier {

    private IDT id;
    private BeanItem<BT> innerItem;
    private final Crud<IDT, BT> crud;
    private final CrudContainer<IDT, BT> container;
    private final List<PropertySetChangeListener> listeners = new ArrayList<PropertySetChangeListener>();
    private boolean writeThrough = false;
    private boolean readThrough = false;
    private boolean modified = false;
    private ChangeHandler changeHandler = new ChangeHandler();


    private class ChangeHandler implements ValueChangeListener {

        @Override
        public void valueChange(ValueChangeEvent event) {
            modified = true;
        }
    }

    public CrudItem(CrudContainer<IDT, BT> container, Crud<IDT, BT> crud, IDT id, BT bean) {
        if(container == null) {
            throw new NullPointerException("Container is null.");
        }
        if(crud == null) {
            throw new NullPointerException("Crud is null.");
        }
        if(bean == null) {
            throw new NullPointerException("Bean is null.");
        }
        this.id = id;
        this.setInnerBean(bean);
        this.crud = crud;
        this.container = container;
    }

    private void reload() {
        if(this.innerItem == null) {
            return;
        }
        BT bean = crud.read(id);
        setInnerBean(bean);
    }

    protected BT getInnerBean() {
        return this.innerItem.getBean();
    }

    protected void setInnerBean(BT bean) {
        if (this.innerItem != null) {
            for(PropertySetChangeListener l : listeners) {
                this.innerItem.removeListener(l);
            }

            for(Object propId : this.getItemPropertyIds()) {
                Property prop = this.innerItem.getItemProperty(propId);
                if(prop instanceof Property.ValueChangeNotifier) {
                    ((Property.ValueChangeNotifier)prop).removeListener(changeHandler);
                }
            }
        }

        this.innerItem = new BeanItem<BT>(bean);
        modified = false;

        PropertySetChangeEvent e = new PropertySetChangeEvent() {
            @Override
            public Item getItem() { return CrudItem.this; }
        };

        if (this.innerItem != null) {

            //Eventlisteners may add og remove themselves from the list. A copy of the list
            //will let us handle that kind of concurrency.
            List<PropertySetChangeListener> copiedListeners = new ArrayList(listeners);
            for(PropertySetChangeListener l : copiedListeners) {
                l.itemPropertySetChange(e);
            }

            for(PropertySetChangeListener l : listeners) {
                this.innerItem.addListener(l);
            }

            for(Object propId : this.getItemPropertyIds()) {
                Property prop = this.innerItem.getItemProperty(propId);
                if(prop instanceof Property.ValueChangeNotifier) {
                    ((Property.ValueChangeNotifier)prop).addListener(changeHandler);
                }
            }
        }



    }

    @Override
    public BT getBean() {
        return innerItem.getBean();
    }

    @Override
    public Property getItemProperty(Object id) {
        return this.innerItem.getItemProperty(id);
    }

    @Override
    public Collection<?> getItemPropertyIds() {
        return this.innerItem.getItemPropertyIds();
    }

    @Override
    public boolean addItemProperty(Object id, Property property) throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Not supported.");
    }

    @Override
    public boolean removeItemProperty(Object id) throws UnsupportedOperationException {
        throw new UnsupportedOperationException("Not supported.");
    }

    @Override
    public void commit() throws SourceException, InvalidValueException {
        if(!(crud instanceof Crud.Editable)) {
            throw new SourceException(this, new UnsupportedOperationException("The crud is not editable."));
        } else {
            Crud.Editable<IDT, BT> editable = (Editable<IDT, BT>) crud;
            editable.update(innerItem.getBean());
            modified = false;
        }
    }

    @Override
    public void discard() throws SourceException {
        if (!modified) {
            return;
        }

        reload();
    }

    @Override
    public boolean isWriteThrough() {
        return writeThrough;
    }

    @Override
    public void setWriteThrough(boolean writeThrough) throws SourceException, InvalidValueException {
        this.writeThrough = writeThrough;
    }

    @Override
    public boolean isReadThrough() {
        return readThrough;
    }

    @Override
    public void setReadThrough(boolean readThrough) throws SourceException {
        this.readThrough = readThrough;
    }

    @Override
    public boolean isModified() {
        return modified;
    }

    @Override
    public void addListener(PropertySetChangeListener listener) {
        innerItem.addListener(listener);
        this.listeners.add(listener);
    }

    @Override
    public void removeListener(PropertySetChangeListener listener) {
        innerItem.removeListener(listener);
        this.listeners.add(listener);
    }


}
